// Copyright 2019 github.com. All rights reserved.
// Use of this source code is governed by github.com.

package util

import (
	"regexp"
	"unicode"
	"unicode/utf8"
)

// SplitPlateNo 分割汽车车牌
func SplitPlateNo(s string) (string, string) {
	idx := 0
	for i, r := range s {
		if !unicode.Is(unicode.Scripts["Han"], r) {
			idx = i
			break
		}
	}
	return s[0:idx], s[idx:]
}

// IsPlateNo 是否为车牌号
func IsPlateNo(s string) bool {
	switch utf8.RuneCountInString(s) {
	case 7:
		return is6lateNo(s)
	case 8:
		return is7PlateNo(s)
	}

	return false
}

// 验证6位车牌号, 即：燃油汽车
func is6lateNo(s string) bool {
	reg := "^[京津冀晋蒙辽吉黑沪苏浙皖闽赣鲁豫鄂湘粤桂琼川贵云渝藏陕甘青宁新使]{1}[A-Z]{1}[0-9A-Z]{5}$"
	matched, err := regexp.MatchString(reg, s)
	if err != nil {
		return false
	}
	if matched {
		bytes := []byte(s)
		count := 0
		for _, c := range bytes {
			if c >= 65 && c <= 90 {
				count++
			}
		}
		if count > 3 {
			return false
		}
		return true
	}
	return false
}

// 验证7位车牌号，即：新能源汽车
func is7PlateNo(s string) bool {
	reg := `^[京津冀晋蒙辽吉黑沪苏浙皖闽赣鲁豫鄂湘粤桂琼川贵云渝藏陕甘青宁新使]{1}[a-zA-Z][dfDF][0-9a-zA-Z]\d{4}$|^[京津冀晋蒙辽吉黑沪苏浙皖闽赣鲁豫鄂湘粤桂琼川贵云渝藏陕甘青宁新使][a-zA-Z]\d{5}[dfDF]$`
	matched, err := regexp.MatchString(reg, s)
	if err != nil {
		return false
	}
	return matched
}

// 车身颜色
var vehicleColor = map[string]string{
	"A": "白",
	"B": "灰",
	"C": "黄",
	"D": "粉",
	"E": "红",
	"F": "紫",
	"G": "绿",
	"H": "蓝",
	"I": "棕",
	"J": "黑",
}

// 车辆类型
var vehicleType = map[string]string{
	"01": "大型汽车",
	"02": "小型汽车",
	"03": "使馆汽车",
	"04": "领馆汽车",
	"05": "境外汽车",
	"06": "外籍汽车",
	"07": "普通摩托车",
	"08": "轻便摩托车",
	"09": "使馆摩托车",
	"10": "领馆摩托车",
	"11": "境外摩托车",
	"12": "外籍摩托车",
	"13": "低速车",
	"14": "拖拉机",
	"15": "挂车",
	"16": "教练汽车",
	"17": "教练摩托车",
	"20": "临时入境汽车",
	"21": "临时入境摩托车",
	"22": "临时行驶车",
	"23": "警用汽车",
	"24": "警用摩托",
	"51": "大型新能源汽车",
	"52": "小型新能源汽车",
}

// 车辆燃油种类
var vehicleFuel = map[string]string{
	"A": "汽油",
	"B": "柴油",
	"C": "电",
	"D": "混合油",
	"E": "天然气",
	"F": "液化石油气",
	"L": "甲醇",
	"M": "乙醇",
	"N": "太阳能",
	"O": "混合动力",
	"P": "氢",
	"Q": "生物燃料",
}

// 使用性质
var vehicleProperty = map[string]string{
	"A": "非营运",
	"B": "公路客运",
	"C": "公交客运",
	"D": "出租客运",
	"E": "旅游客运",
	"F": "货运",
	"G": "租赁",
	"H": "警用",
	"I": "消防",
	"J": "救护",
	"K": "工程救险",
	"L": "营转非",
	"M": "出租转非",
	"N": "教练",
	"O": "幼儿校车",
	"P": "小学生校车",
	"Q": "初中生校车",
	"R": "危化品运输",
	"S": "中小学生校车",
}

// VehicleColor 映射车辆车身颜色
//   code 颜色编号 A => 白色
func VehicleColor(code string) string {
	if v, ok := vehicleColor[code]; ok {
		return v
	}

	return ""
}

// VehicleType 映射车辆类型
//   code 类型编号 02 = 小型车
func VehicleType(code string) string {
	if v, ok := vehicleType[code]; ok {
		return v
	}

	return ""
}

// VehicleType 映射车辆燃油种类
//   code 类型编号 A => 汽油
func VehicleFuel(code string) string {
	if v, ok := vehicleFuel[code]; ok {
		return v
	}

	return ""
}

// VehicleType 映射车辆使用性质
//   code 类型编号 A => 非营运
func VehicleProperty(code string) string {
	if v, ok := vehicleProperty[code]; ok {
		return v
	}

	return ""
}
